/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.generator.service;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import net.sf.parser4j.generator.entity.grammardefnode.EnumStatementDefType;
import net.sf.parser4j.generator.entity.grammardefnode.GrammarDef;
import net.sf.parser4j.generator.entity.grammardefnode.IGrammarDefNode;
import net.sf.parser4j.generator.entity.grammardefnode.RuleDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesDef;
import net.sf.parser4j.generator.entity.grammardefnode.RulesGroupListDef;
import net.sf.parser4j.generator.entity.grammardefnode.StatementDef;
import net.sf.parser4j.generator.service.grammardefnode.IGrammarDefVisitor;
import net.sf.parser4j.kernelgenerator.entity.GrammarTreeFactoryData;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.AbstractMularyGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.GrammarNodeNull;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.IGrammarNode;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeConcat;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeString;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.GrammarNodeTerminalSetOfTerminal;
import net.sf.parser4j.kernelgenerator.entity.grammarnode.impl.OperandGrammarNode;
import net.sf.parser4j.kernelgenerator.service.GeneratorException;
import net.sf.parser4j.parser.entity.EnumNodeType;
import net.sf.parser4j.parser.entity.data.NonTerminalMap;
import net.sf.parser4j.parser.entity.data.Pair;

/**
 * create grammar from root data node
 * 
 * @author luc peuvrier
 * 
 */
public final class GrammarFactory implements IGrammarDefVisitor {

	private static final String START_NON_TERMINAL = "Start";

	private static final String EMPTY_STRING = "";

	private static final GrammarFactory INSTANCE =
	/**/new GrammarFactory();

	private static final StatementToGrammarNode statementToGrammarNode =
	/**/StatementToGrammarNode.getInstance();

	private transient NonTerminalMap nonTerminalMap;

	private transient Map<String, Pair<StatementDef, IGrammarNode>>
	/**/statementDefAnGrammarNodeMap;

	/** white space non terminal name for group of rules, empty string if none */
	private transient String whiteSpaceNonTerminalName;

	public static GrammarFactory getInstance() {
		return INSTANCE;
	}

	private GrammarFactory() {
		super();
	}

	/**
	 * create grammar tree
	 * 
	 * @param rootData
	 *            root data node of grammar definition text file parsing
	 * @return
	 * @throws GeneratorException
	 */
	public GrammarTreeFactoryData createGrammar(final IGrammarDefNode rootData)
			throws GeneratorException {
		nonTerminalMap = new NonTerminalMap();
		whiteSpaceNonTerminalName = null;// NOPMD nothing defined
		/* will be populate by visit */
		statementDefAnGrammarNodeMap =
		/**/new TreeMap<String, Pair<StatementDef, IGrammarNode>>();
		try {
			/* do the visit */
			rootData.accept(this);
		} catch (GrammarDefNodeVisitException exception) {
			throw new GeneratorException(exception);
		}
		linkGrammarNode();
		// FIXME Start must exist
		// FIXME Start must not be referenced
		// FIXME warn unused
		Pair<StatementDef, IGrammarNode> pair = statementDefAnGrammarNodeMap
				.get(START_NON_TERMINAL);
		IGrammarNode rootGrammarNode = pair.getMemberB();
		/*
		 * special case of rule: Start : X ; where X is a non terminal
		 */
		if (EnumNodeType.NULL.equals(rootGrammarNode.getGrammarNodeType())) {
			final StatementDef statementDef = pair.getMemberA();
			final IGrammarNode whiteSpaceGrammarNode =
			/**/getWhiteSpaceGrammarNode(statementDef);
			final String name = statementDef.getNonTerminalName();
			final IGrammarNode subConcatGrammarNode = statementDefAnGrammarNodeMap
					.get(name).getMemberB();
			rootGrammarNode = createConcatNode(
					statementDef.getMatchClassName(), statementDef,
					subConcatGrammarNode, whiteSpaceGrammarNode, true);

			pair = new Pair<StatementDef, IGrammarNode>(statementDef,
					rootGrammarNode);
			statementDefAnGrammarNodeMap.put(START_NON_TERMINAL, pair);
		}
		if (rootGrammarNode == null) {
			throw new GeneratorException("Start non terminal is not defined");
		}
		return new GrammarTreeFactoryData(nonTerminalMap, rootGrammarNode);
	}

	private void linkGrammarNode() throws GeneratorException {
		for (Pair<StatementDef, IGrammarNode> pair : statementDefAnGrammarNodeMap
				.values()) {
			final StatementDef statementDef = pair.getMemberA();
			final IGrammarNode grammarNode = pair.getMemberB();
			final EnumStatementDefType statementDefType = statementDef
					.getStatementDefType();
			final EnumNodeType grammarNodeType = grammarNode
					.getGrammarNodeType();
			switch (grammarNodeType) {
			case ALTERNATIVE:
				if (EnumStatementDefType.ALTERNATIVE.equals(statementDefType)) {
					linkGrammarNode(grammarNodeType, statementDef,
							(AbstractMularyGrammarNode) grammarNode);
				}
				break;
			case CONCAT:
				if (EnumStatementDefType.CONCAT.equals(statementDefType)) {
					linkGrammarNode(grammarNodeType, statementDef,
							(AbstractMularyGrammarNode) grammarNode);
				} else if (EnumStatementDefType.NON_TERMINAL
						.equals(statementDefType)) {
					final String name = statementDef.getNonTerminalName();
					final List<OperandGrammarNode> operandList =
					/**/new LinkedList<OperandGrammarNode>();// NOPMD
					final IGrammarNode nonTerminalGrammarNode =
					/**/statementDefAnGrammarNodeMap.get(name).getMemberB();
					final IGrammarNode whiteSpaceGrammarNode =
					/**/getWhiteSpaceGrammarNode(statementDef);
					if (whiteSpaceGrammarNode != null
							&& (nonTerminalGrammarNode.isTerminal() || nonTerminalGrammarNode
									.isToken())) {
						addToOperandList(operandList, whiteSpaceGrammarNode,
								true);
					}
					addToOperandList(operandList, nonTerminalGrammarNode, false);
					/* add white space at end of concat start rules */
					if (whiteSpaceGrammarNode != null
							&& START_NON_TERMINAL.equals(grammarNode
									.getNonTerminalName())) {
						addToOperandList(operandList, whiteSpaceGrammarNode,
								true);
					}
					((AbstractMularyGrammarNode) grammarNode)
							.setOperandList(operandList);
				}
				break;
			case EMPTY:
				/* have not son(s) */
				break;
			case CHARACTER_TERMINAL:
				if (EnumStatementDefType.TERMINAL_SET_OF_TERMINAL
						.equals(statementDefType)) {
					final List<StatementDef> list = statementDef
							.getSetOfTerminal();
					final Set<IGrammarNode> setOfTerminalNode =
					/**/new TreeSet<IGrammarNode>();// NOPMD
					for (StatementDef subs : list) {
						final String name = subs.getNonTerminalDefinedName();
						setOfTerminalNode.add(statementDefAnGrammarNodeMap.get(
								name).getMemberB());
					}
					((GrammarNodeTerminalSetOfTerminal) grammarNode)
							.setSetOfTerminalNode(setOfTerminalNode);

				}
				break;
			case STRING_TERMINAL:
				final String stringValue = statementDef.getTerminalString();
				((GrammarNodeString) grammarNode).setStringValue(stringValue,
						nonTerminalMap);
				break;
			case NULL:
				/* have not son(s), identifier statement special case */
				break;
			default:
				throw new GeneratorException("bad type " + grammarNodeType);
			}
		}
	}

	private void linkGrammarNode(final EnumNodeType grammarNodeType,
			final StatementDef statementDef,
			final AbstractMularyGrammarNode grammarNode)
			throws GeneratorException {

		/* true if concat node, else must be alternative node */
		final boolean concat = EnumNodeType.CONCAT.equals(grammarNodeType);

		final IGrammarNode whiteSpaceGrammarNode =
		/**/getWhiteSpaceGrammarNode(statementDef);

		final List<OperandGrammarNode> operandList =
		/**/new LinkedList<OperandGrammarNode>();
		final List<StatementDef> list = statementDef.getOperandList();
		for (StatementDef subStatementDef : list) {
			String name = subStatementDef.getNonTerminalDefinedName();
			IGrammarNode subGrammarNode = statementDefAnGrammarNodeMap
					.get(name).getMemberB();
			final String matchClassName = subStatementDef.getMatchClassName();
			/*
			 * non terminal special case
			 */
			if (EnumNodeType.NULL.equals(subGrammarNode.getGrammarNodeType())) {
				if (EnumStatementDefType.NON_TERMINAL.equals(subStatementDef
						.getStatementDefType())) {
					/*
					 * asserted statement type is a non terminal
					 */

					if (matchClassName == null) {
						/*
						 * get the grammar node for the non terminal
						 */
						name = subStatementDef.getNonTerminalName();
						subGrammarNode = statementDefAnGrammarNodeMap.get(name)
								.getMemberB();
					} else {
						/*
						 * get the grammar node for the non terminal
						 */
						name = subStatementDef.getNonTerminalName();
						final IGrammarNode subConcatGrammarNode =
						/**/statementDefAnGrammarNodeMap.get(name)
								.getMemberB();
						/*
						 * create the grammar node for the non terminal
						 */
						subGrammarNode = createConcatNode(matchClassName,
								subStatementDef, subConcatGrammarNode,
								null/* whiteSpaceGrammarNode */, false);
					}
				} else {
					throw new GeneratorException("bad type "
							+ subStatementDef.getStatementDefType());
				}
			}

			if (concat
					&& whiteSpaceGrammarNode != null
					&& (subGrammarNode.isTerminal() || subGrammarNode.isToken())) {
				/* additional white space son */
				addToOperandList(operandList, whiteSpaceGrammarNode, true);
			}
			/* add son */
			addToOperandList(operandList, subGrammarNode, false);
		}
		/* add white space at end of concat start rules */
		if (concat && whiteSpaceGrammarNode != null
				&& START_NON_TERMINAL.equals(grammarNode.getNonTerminalName())) {
			/* additional white space son */
			addToOperandList(operandList, whiteSpaceGrammarNode, true);
		}
		grammarNode.setOperandList(operandList);
	}

	/**
	 * Insert white space before son grammar node if the son grammar node is a
	 * terminal or a token and white space grammar node is defined<br>
	 * Also optionally add ending white space<br>
	 * 
	 * @param matchClassName
	 *            match manager class name of concat grammar node to create,
	 *            empty string if none
	 * @param statementDef
	 *            statement definition to use for grammar concat node creation
	 * @param sonGrammarNode
	 *            son grammar node of created concat grammar node
	 * @param whiteSpaceGrammarNode
	 *            the white space grammar node, null if no white space to insert
	 * @param addEndingWhiteSpace
	 *            true if add ending white space
	 * @return
	 * @throws GeneratorException
	 */
	private IGrammarNode createConcatNode(final String matchClassName,
			final StatementDef statementDef, final IGrammarNode sonGrammarNode,
			final IGrammarNode whiteSpaceGrammarNode,
			final boolean addEndingWhiteSpace) throws GeneratorException {
		final GrammarNodeConcat grammarNode = statementToGrammarNode
				.reCreateConcat(statementDef, nonTerminalMap);
		final List<OperandGrammarNode> operandList =
		/**/new LinkedList<OperandGrammarNode>();
		if (whiteSpaceGrammarNode != null
				&& (sonGrammarNode.isTerminal() || sonGrammarNode.isToken())) {
			/* additional white space son */
			addToOperandList(operandList, whiteSpaceGrammarNode, true);
		}
		addToOperandList(operandList, sonGrammarNode, false);
		if (addEndingWhiteSpace && whiteSpaceGrammarNode != null) {
			/* additional ending white space son */
			addToOperandList(operandList, whiteSpaceGrammarNode, true);
		}
		grammarNode.setOperandList(operandList);
		grammarNode.setMatchClassSimpleName(matchClassName);
		return grammarNode;
	}

	private void addToOperandList(final List<OperandGrammarNode> operandList,
			final IGrammarNode grammarNode, final boolean whiteSpace) {
		if (EnumNodeType.NULL.equals(grammarNode.getGrammarNodeType())) {
			throw new IllegalArgumentException(
					"can not add null grammar node to operand list");
		}
		final OperandGrammarNode operandGrammarNode = new OperandGrammarNode();
		operandGrammarNode.setGrammarNode(grammarNode);
		operandGrammarNode.setWhiteSpace(whiteSpace);
		operandList.add(operandGrammarNode);
	}

	private IGrammarNode getWhiteSpaceGrammarNode(final StatementDef def) {
		whiteSpaceNonTerminalName = def.getWhiteSpaceNonTerminalName();
		if (whiteSpaceNonTerminalName == null) {
			throw new IllegalStateException(
					"white space non terminal value must be set for " + def);
		}
		final IGrammarNode whiteSpaceGrammarNode;
		if (EMPTY_STRING.equals(whiteSpaceNonTerminalName)) {
			whiteSpaceGrammarNode = null;
		} else {
			whiteSpaceGrammarNode = statementDefAnGrammarNodeMap.get(
					whiteSpaceNonTerminalName).getMemberB();
		}
		return whiteSpaceGrammarNode;
	}

	@Override
	public void beginVisit(final GrammarDef def) {
		// nothing to implement
	}

	@Override
	public void endVisit(final GrammarDef def) {
		// nothing to implement
	}

	@Override
	public void beginVisit(final RulesGroupListDef def) {
		// nothing to implement
	}

	@Override
	public void endVisit(final RulesGroupListDef def) {
		// nothing to implement
	}

	@Override
	public void beginVisit(final RulesDef def) {
		whiteSpaceNonTerminalName = def.getWhiteSpaceNonTerminalName();
	}

	@Override
	public void endVisit(final RulesDef def) {
		whiteSpaceNonTerminalName = null;// NOPMD
	}

	@Override
	public void beginVisit(final RuleDef def) {
		// nothing to implement
	}

	@Override
	public void endVisit(final RuleDef def) {
		// nothing to implement
	}

	@Override
	public boolean beginVisit(final StatementDef statementDef)
			throws GrammarDefNodeVisitException {
		final boolean visit;
		if (whiteSpaceNonTerminalName == null) {
			throw new IllegalStateException(
					"white space non terminal must be defined");
		}
		statementDef.setWhiteSpaceNonTerminalName(whiteSpaceNonTerminalName);
		final String nonTerminalDefinedName = statementDef
				.getNonTerminalDefinedName();
		try {
			final IGrammarNode grammarNode = statementToGrammarNode.create(
					statementDef, nonTerminalMap);

			if (!EMPTY_STRING.equals(whiteSpaceNonTerminalName)
					&& !GrammarNodeNull.class.equals(grammarNode.getClass())
					&& (grammarNode.isTerminal() || grammarNode.isToken())) {

				final String subNonTerminalDefinedName = "_"
						+ nonTerminalDefinedName;
				final StatementDef statementDefCopy = new StatementDef(
						statementDef, subNonTerminalDefinedName);

				statementDef.changeToConcat(statementDefCopy);
				final GrammarNodeConcat newGrammarNode =
				/**/statementToGrammarNode.reCreateConcat(statementDef,
						nonTerminalMap);
				addToStatementDefAnGrammarNodeMap(nonTerminalDefinedName,
						statementDef, newGrammarNode);

				grammarNode.setNonTerminalName(subNonTerminalDefinedName);
				final int nonTerminalIdentifier = nonTerminalMap
						.createNonTerminalIdentifier(subNonTerminalDefinedName,
								null/* matchClassName */, statementDefCopy
										.isToken());
				grammarNode.setNonTerminalIdentifier(nonTerminalIdentifier);
				addToStatementDefAnGrammarNodeMap(subNonTerminalDefinedName,
						statementDefCopy, grammarNode);
				visit = false;
			} else {
				addToStatementDefAnGrammarNodeMap(nonTerminalDefinedName,
						statementDef, grammarNode);
				visit = true;
			}
		} catch (GeneratorException exception) {
			throw new GrammarDefNodeVisitException(exception);
		}
		return visit;
	}

	private void addToStatementDefAnGrammarNodeMap(
			final String nonTerminalDefinedName,
			final StatementDef statementDef, final IGrammarNode grammarNode)
			throws GrammarDefNodeVisitException {
		final Pair<StatementDef, IGrammarNode> value =
		/**/new Pair<StatementDef, IGrammarNode>(statementDef, grammarNode);
		if (statementDefAnGrammarNodeMap.put(nonTerminalDefinedName, value)
		/**/!= null) {
			throw new GrammarDefNodeVisitException("duplicate definition of "
					+ nonTerminalDefinedName);
		}
	}

	@Override
	public void endVisit(final StatementDef statementDef) {
		// nothing to implement
	}
}
