/*
 * Copyright 2008 Luc Peuvrier
 * 
 * This file is part of parser4j.
 *
 * parser4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * parser4j is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with parser4j.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.parser4j.parser.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;

import org.apache.log4j.Logger;

/**
 * 
 * @author luc peuvrier
 * 
 */
public class ParserInput {

	private static final Logger _log = Logger.getLogger(ParserInput.class);

	private final transient BufferedReader bufferedReader;

	private transient boolean openned;

	private transient int lineNumber;

	private transient int columnNumber;

	private final transient Deque<Integer> unreadQue = new LinkedList<Integer>();

	private final transient Deque<Integer> markQue = new LinkedList<Integer>();

	private transient boolean markFlag = false;

	private int markLineNumber; // NOPMD

	private int markColumnNumber;// NOPMD

	public ParserInput(final Reader reader) throws ParserException {
		super();
		bufferedReader = new BufferedReader(reader);
		lineNumber = 1;
		columnNumber = 1;
		openned = true;
	}

	public int read() throws IOException {
		if (!openned) {
			throw new IOException("closed");
		}
		int readed;
		if (unreadQue.isEmpty()) {
			try {
				readed = bufferedReader.read();
			} catch (IOException exception) {
				try {
					bufferedReader.close();
				} catch (IOException exception2) {// NOPMD
					// ignore error
				}
				openned = false;
				throw exception;
			}
		} else {
			readed = unreadQue.pollFirst();
		}
		if (markFlag && readed != -1) {
			markQue.addLast(readed);
		}
		if (readed == '\n') {
			lineNumber++;
			columnNumber = 1;
		} else {
			columnNumber++;
		}
		if (_log.isDebugEnabled()) {
			if (readed == -1) {
				_log.debug("readed eof");
			} else {
				_log.debug("readed " + (char) readed);
			}
		}
		return readed;
	}

	public void close() {
		if (openned) {
			try {
				bufferedReader.close();
			} catch (IOException exception) {
				_log.error("closing reader", exception);
			}
			openned = false;
		}
	}

	public int getColumnNumber() {
		return columnNumber;
	}

	public int getLineNumber() {
		return lineNumber;
	}

	/**
	 * Marks the present position in the input. Subsequent calls to
	 * {@link #reset()} will attempt to reposition the input to this point.
	 */
	public void mark() {
		markQue.clear();
		markFlag = true;
		markColumnNumber = columnNumber;
		markLineNumber = lineNumber;
	}

	/**
	 * Resets the input to the most recent mark.
	 */
	public void reset() {
		markFlag = false;
		final Iterator<Integer> iterator = markQue.descendingIterator();
		while (iterator.hasNext()) {
			unreadQue.addFirst(iterator.next());
		}
		columnNumber = markColumnNumber;
		lineNumber = markLineNumber;
	}
}
